#!/usr/bin/env python
# encoding: utf-8
"""
projector.py

Project an SBML model, based on a scaffold model and set of .rel files

Created by Nicolas Loira on 2010-12-28.
Copyright (c) 2010 LaBRI, Univ.Bordeaux I, France. All rights reserved.
"""

import sys
import getopt
from .pathtastictools import SBMLmodel, MultiRelations


HELP_MESSAGE = '''
Project an SBML file from a scaffold model and a set or .rel files
usage: projector.py [-f curated.rel] scaffold.xml method1.rel [method2.rel...]
or print the current version id with: projector -v
'''

VERSION = "0.1.2"


def log(message):
	"""Print to stderr"""
	print >>sys.stderr, message

class Usage(Exception):
	"""Exception in case of demand of help or error parsing parameters"""
	def __init__(self, msg):
		super(Usage, self).__init__()
		self.msg = msg

def getForcedReactions(forceGAfile):
	"""check if we are forcing reactions and parse that file"""
	forcedReactions = None
	if forceGAfile:
		forcedReactions = dict()
		for line in open(forceGAfile):
			if line.startswith('#'):
				continue
			rid, newga = line[:-1].split("\t")
			forcedReactions[rid] = newga
			log("forcing: %s to (%s)" % (rid, newga))
	return forcedReactions


def parseParameters(argv):
	"""Parse parameters from command line"""
	forceGAfile = None
	output = None
	logFile = None

	if argv is None:
		argv = sys.argv
	try:
		try:
			opts, args = getopt.getopt(argv[1:], "ho:vf:l:", ["help", "output=", "force=", "log="])
		except getopt.error, msg:
			raise Usage(msg)

		# option processing
		for option, value in opts:
			if option == "-v":
				print VERSION
				sys.exit()
			if option in ("-h", "--help"):
				raise Usage(HELP_MESSAGE)
			if option in ("-o", "--output"):
				output = value
			if option in ("-f", "--force"):
				forceGAfile = value
			if option in ("-l", "--log"):
				logFile = value


	except Usage, err:
		print >> sys.stderr, sys.argv[0].split("/")[-1] + ": " + str(err.msg)
		print >> sys.stderr, "\t for help use --help"
		sys.exit()

	return args, forceGAfile, output, logFile

def main(argv=None):
	"""Projector component of Pantograph. Takes a model and orthology files to produce a new model"""

	args, forceGAfile, output, logFile = parseParameters(argv)

	# check parameters
	# parse filenames from args
	if len(args) < 2:
		log(HELP_MESSAGE)
		return 2

	sbmlfile = args[0]
	relFiles = args[1:]

	# check if we are forcing reactions
	forcedReactions = getForcedReactions(forceGAfile)

	# logfile or stderr
	logFD = sys.stderr if logFile is None else open(logFile, 'w')

	# parse SBML
	model = SBMLmodel(sbmlfile, logFD)

	# a Relations class for each .rel file
	rels = MultiRelations(logFD=logFD)
	rels.loadMulti(relFiles)

	# DEBUG
	# print "DEBBUG: Checking rels"
	# rels.debugRels()

	# start unraveling among rels only

	rels.unravel1to1()
	rels.unravelDuplicates()

	# report findings
#	for k,v in rels.map1to1.iteritems():
#		print ">>>%s --> %s" % (str(k),str(v))
#	for k,v in rels.map1toN.iteritems():
#		print "@@@%s --> %s" % (str(k),str(v))

	# force new GAs to certain reactions
	if forceGAfile:
		model.forceGeneAssociations(forcedReactions)

	# rewrite GeneAssociations, using rels.unraveling and GAs.unraveling
	model.rewriteGeneAssociations2(rels)



	# clean resulting model
	model.removeReactions()

	# output projected model
	outfd = sys.stdout if output is None else open(output, "w")
	model.write(outfd)



if __name__ == "__main__":
	sys.exit(main())



#!/usr/bin/env python

"""
omcl2rel.py


Parses OrthoMCL output and translate it to a simple .rel file

Created by Nicolas Loira on 2010-06-15.
Copyright (c) 2010 LaBRI, Univ.Bordeaux I, France. All rights reserved.
"""


import sys
from collections import defaultdict
import getopt

VERSION="0.1.2"




#### params
def main():


	def printrel(genes):
		if len(genes[source])>0 or len(genes[target])>0:
			outfd.write(",".join( [rename.get(g,g) for g in genes[target]]) + \
				"\t" + \
				",".join( [rename.get(g,g) for g in genes[source]]) + "\n")


	output = None
	corrFile = None
	rename=dict()

	format=None

	source = 'scer'
	target = 'ylip'

	targetTranslator={
	'CAGL':'cgla',
	'DEHA':'dhan',
	'KLLA':'klac',
	'YALI':'ylip'
	}

	opts, args = getopt.getopt(sys.argv[1:], "ho:vs:t:2", ["help", "output=", "corr=", "source=", "target=","format="])

	for o, v in opts:
		if o == "--corr": corrFile=v
		if o in ("-s","--source"): source=v
		if o in ("-t","--target"): target=v
		if o in ("--format"): format=v
		if o == "-2": format="format2"
		if o == "-v":
			print VERSION
			return 0
		if o in ("-o","--output"):
			output = v



	groupfile = args[0]

	# output to file?
	if output is not None:
		outfd = open(output,"w")
	else:
		outfd = sys.stdout


	# try to guess format
	if format==None:
		groupsFD=open(groupfile)
		l=groupsFD.readline()
		if l.find("\t")>-1:
			format="format2"
		else:
			format="format1"
		groupsFD.close()



	# rename target to omcl standard
	if target in targetTranslator:
		target=targetTranslator[target]

	### Correlation

	if corrFile:
		for line in open(corrFile):
			elems=line[:-1].split()
			rename[elems[0]]=elems[1]


	#### parse groups

	if format=="format1":
		for line in open(groupfile):
			elems = line[:-1].split()
			genes=defaultdict(list)
			for e in elems[1:]:
				parts = e.split("|")
				genes[parts[0]].append(parts[1])
			printrel(genes)
	elif format=="format2":
		for line in open(groupfile):
			elems = line[:-1].split('\t')
			elems = elems[1].split()
			genes=defaultdict(list)
			for e in elems:
				parts = e[:-1].split("(")
				# print parts[0],parts[1]
				genes[parts[1]].append(parts[0])
			printrel(genes)
	else:
		assert "No default format?"

	# closing
	outfd.close()


	#	if len(genes['ylip'])>0 or len(genes['scer'])>0:
	#		print ",".join(genes['ylip'])+"\t"+",".join(genes['scer'])
		# for omcl
		
		# use the name in omcl in case we cannot translate it (but remove the 'xxxx|' orgID at the begginig)


if __name__ == '__main__':
    main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import libsbml

"""High level SBML utils over libsbml"""

__author__ = "Nicolas Loira"
__email__ = "nloira@gmail.com"
__date__ = "14/March/2016"


def read_sbml(sbmlfile):
	document = libsbml.readSBML(sbmlfile)
	errors = document.getNumErrors()

	if errors > 0:
		document.printErrors()
		sys.exit("Invalid SBML.")
	else:
		return document


def log(msg):
	"""Write to stderr"""
	sys.stderr.write(msg + '\n')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import libsbml
from collections import defaultdict

import BiGG as bigg

"""bigg2sbml: add bigg elements to an existing sbml file"""

__author__ = "Nicolas Loira"
__email__ = "nloira@gmail.com"
__date__ = "24/July/2015"



def get_full_species_id(m):
	return str('M_' + m['bigg_id'] + '_' + m['compartment_bigg_id'])


def add_species(model, m):

	assert model
	assert m

	m_id = get_full_species_id(m)

	# check compartments
	compartment = str(m['compartment_bigg_id'])
	if model.getCompartment(compartment) is None:
		# add_compartment(model, compartment)		# DON'T!
		# move it to cytosol
		m_id = m_id[:-2] + "_c"
		compartment = "c"

		renamed_species = model.getSpecies(m_id)
		if renamed_species is not None:
			return renamed_species

	s = model.createSpecies()
	s.setCompartment(compartment)
	s.setId(m_id)
	s.setName(str(m.get('name', 'NO_NAME')))

	# let's ask bigg for additional info
	bigg_species = bigg.get_bigg_species(m['bigg_id'])

	species_notes = ""

	if 'formulae' in bigg_species:
		species_notes = ""
		for f in bigg_species['formulae']:
			species_notes += "<html:p>FORMULA: " + str(f) + "</html:p>"

		body = "<html:body>" + species_notes + "</html:body>"
		s.setNotes(body)

	return s


def add_compartment(model, compartment):
	if compartment in bigg.BIGG_COMPARTMENTS:
		newc = model.createCompartment()
		newc.setId(compartment)
		newc.setName(bigg.BIGG_COMPARTMENTS[compartment])


def add_parameter(kl, id, value):
	p = kl.createParameter()
	p.setId(id)
	p.setValue(value)
	p.setUnits("mmol_per_gDW_per_hr")
	return p


def get_reaction_results_from_any_model(biggreaction):
	"""From this generic reaction, get the reaction_results from any of the models where this reaction appears"""

	if 'models_containing_reaction' not in biggreaction:
		return None

	anymodel_id = biggreaction['models_containing_reaction'][0]['bigg_id']
	anyreaction = bigg.get_bigg_reaction_model(biggreaction['bigg_id'], anymodel_id)

	if 'results' in anyreaction:
		return anyreaction['results'][0]
	else:
		return None


def add_reaction(model, biggreaction, gene_association=None):

	assert model is not None
	assert biggreaction is not None
	subsystem = None

	# create reaction
	r = model.createReaction()
	r.setId("R_" + str(biggreaction['bigg_id']))
	r.setName(str(biggreaction['name']))

	# inherit constraints from model
	if 'results' in biggreaction:
		reaction_results = biggreaction['results'][0]
	else:
		reaction_results = get_reaction_results_from_any_model(biggreaction)

	if reaction_results is not None:
		lb = reaction_results['lower_bound']
		ub = reaction_results['upper_bound']
		# oc = reaction_results['objective_coefficient']
		subsystem = reaction_results.get('subsystem', None)

		r.setReversible(lb < 0)

		kl = r.createKineticLaw()
		add_parameter(kl, 'LOWER_BOUND', lb)
		add_parameter(kl, 'UPPER_BOUND', ub)
		add_parameter(kl, 'OBJECTIVE_COEFFICIENT', 0.0)
		add_parameter(kl, 'FLUX_VALUE', 0.0)
		kl.setFormula(" FLUX_VALUE ")

	# add new metabolites from reactant/products
	for m in biggreaction['metabolites']:

		m_id = get_full_species_id(m)
		species = model.getSpecies(m_id)
		if species is None:
			species = add_species(model, m)

		stoichiometry = m['stoichiometry']
		if stoichiometry > 0:
			s = r.createReactant()
		else:
			s = r.createProduct()

		s.setStoichiometry(abs(stoichiometry))
		s.setSpecies(species.getId())

	# add gene association and subsystem
	notes = ""
	if gene_association is not None:
		notes += "<html:p>GENE_ASSOCIATION: " + gene_association + "</html:p>"

	if subsystem is not None:
		notes += "<html:p>SUBSYSTEM: " + subsystem + "</html:p>"

	# add notes to reaction
	if notes != "":
		body = "<html:body>" + str(notes) + "</html:body>"
		r.setNotes(body)

	return r





#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import

import sys
import itertools
from pprint import pprint

import libsbml

from .SBMLutils import log

MAX_FLOW_BOUND = 999999

"""High level model edit operations/verbs over libsbml"""

__author__ = "Nicolas Loira"
__email__ = "nloira@gmail.com"
__date__ = "14/March/2016"

already_moved = dict()

def valid_compartment(model, compartment):
	compartments = [c.getId() for c in model.getListOfCompartments()]
	return compartment in compartments


def move_reaction(model, reactionid, source_comp, target_comp):
	"Move species associated to this reaction, from source compartment to target compartment. Create species if necessary"
	assert model is not None

	if not valid_compartment(model, source_comp) or not valid_compartment(model, target_comp):
		log("MOVE: invalid compartments (%s,%s)" % (source_comp, target_comp))
		return False

	reaction = model.getReaction(reactionid)
	speciesReferences = itertools.chain(reaction.getListOfProducts(), reaction.getListOfReactants())
	
	for sr in speciesReferences:
		sid = sr.getSpecies()
		species = model.getSpecies(sid)
		if species is None:
			sr.setSpecies(already_moved[sid])  # probably already moved
		elif species.getCompartment() == source_comp:
			log("Moving %s from [%s] to [%s]" % (species.getId(), source_comp, target_comp))
			new_speciesid = move_species(model, sid, source_comp, target_comp)
			sr.setSpecies(new_speciesid)

	return True


def move_species(model, speciesid, source_comp, target_comp):
	global already_moved

	if speciesid in already_moved:
		return already_moved[speciesid]

	species = model.getSpecies(speciesid)
	assert species.getCompartment() == source_comp

	# if there is an existing species at the target compartment, use that one
	# else, create a new one with the same characteristics

	new_speciesid = speciesid[:-1] + target_comp
	
	if model.getSpecies(new_speciesid) is None:
		new_species = species.clone()
		new_species.setCompartment(target_comp)
		new_species.setId(new_speciesid)
		log(" Cloning %s to %s" % (speciesid, new_species.getId()))
		model.addSpecies(new_species)
		model.removeSpecies(speciesid)
	else:
		log(" Using existing "+new_speciesid)

	already_moved[speciesid] = new_speciesid
	return new_speciesid


def remove_compartment(model, compartmentid):
	
	# remove all species associated to this compartment
	for species in list(model.getListOfSpecies()):
		if species.getCompartment() == compartmentid:
			model.removeSpecies(species.getId())

	# remove all related reactions
	reactionids = reactions_touching_compartment(model, compartmentid)
	for r in reactionids:
		model.removeReaction(r)

	# now remove the compartment
	model.removeCompartment(compartmentid)

def reactions_touching_compartment(model, compartmentid):

	csufix = "_"+compartmentid
	touching = []

	for reaction in model.getListOfReactions():
		speciesReferences = itertools.chain(reaction.getListOfProducts(), reaction.getListOfReactants())
		inCompartment = [s for s in speciesReferences if s.getSpecies().endswith(csufix)]
		if len(inCompartment) > 0:
			touching.append(reaction.getId())

	return touching

def remove_requirement(model, reactionid, speciesid):
	reaction = model.getReaction(reactionid)
	reaction.removeReactant(speciesid)


def set_standard_size(model, sizev=1.0):
	assert model

	for c in model.getListOfCompartments():
		if not c.isSetSize():
			c.setSize(sizev)


def set_species_initial_concentration(model, initial=0.0):
	assert model

	for s in model.getListOfSpecies():
		if not s.isSetInitialConcentration():
			s.setInitialConcentration(initial)


def set_gene_association(model, reactionid, gene_association):
	assert model

	reaction = model.getReaction(reactionid)
	if reaction is None:
		log("Unable to set GENE_ASSOCIATION to:"+reactionid)
	else:
		remove_gene_association(reaction)
		reaction.appendNotes("<html:p>GENE_ASSOCIATION: " + gene_association + "</html:p>")

	return


def remove_gene_association(reaction):
	if not reaction.isSetNotes():
		return

	lines = reaction.getNotesString().split("\n")
	filterGA = [l for l in lines if l.find("GENE_ASSOCIATION") == -1]
	reaction.setNotes("\n".join(filterGA))

def remove_orphan_species(model):
	removed = []
	for species in list(model.getListOfSpecies()):
		if model.getSpeciesReference(species.getId()) is None:
			removed.append(species.getId())
			species.removeFromParentAndDelete()

	return removed if len(removed) > 0 else None

def set_reversibility(model, reactionid, reversible=True):
	reaction = model.getReaction(reactionid)
	if reaction is None:
		log("Unable to set reversibility to:"+reactionid)
		return

	reaction.setReversible(reversible)

	kl = reaction.getKineticLaw()
	kl.getParameter('LOWER_BOUND').setValue(-MAX_FLOW_BOUND if reversible else 0)
	kl.getParameter('UPPER_BOUND').setValue(MAX_FLOW_BOUND)




#!/usr/bin/env python
# encoding: utf-8
"""
inparanoid2rel.py

Parses inparanoid output and translate it to a simple .rel file

Created by Nicolas Loira on 2010-06-15.
Copyright (c) 2010 LaBRI, Univ.Bordeaux I, France. All rights reserved.
"""

import sys
import os
import getopt

from xml.etree.ElementTree import *
from collections import defaultdict

import re

VERSION="0.1.2"


def main():
	output = None

	# parse parameters
	corrFile=None
	opts, args = getopt.getopt(sys.argv[1:], "ho:vs:t:", ["help", "output=", "corr="])
	for o, v in opts:
		if o == "--corr": corrFile=v
		if o == "-v":
			print VERSION
			return 0
		if o in ("-o","--output"):
			output = v

	
	# parse correlation file, if it exists
	rename=dict()
	if corrFile:
		for line in open(corrFile):
			elems=line[:-1].split()
			rename[elems[0]]=elems[1]
	
	# output to file?
	if output is not None:
		outfd = open(output,"w")
	else:
		outfd = sys.stdout
	
	# infile is first non-named parameter
	
	infile = args[0]

	species = None

	# try xml parser first
	if infile.lower().find("xml")>-1:
	    species, groups = parseInparanoidXML(infile)
	else:
		groups=parseInparanoidTable(infile)

	order=None
	iFirst=None
	iSecond=None
	
	saceLocus=re.compile("Y[A-Z][RL]\d{3}[WC]")
	
	for g in groups:
		if order is None:
			# which one is sace?
			if saceLocus.match(g[0][0]):
				iFirst, iSecond= 1,0
			else:
				iFirst, iSecond= 0,1
			
		# first target, then sace
		# ugly, I know
		gFirst=[rename.get(x,x) for x in g[iFirst]]
		gSecond=[rename.get(x,x) for x in g[iSecond]]		
		line = ",".join(gFirst)+"\t"+",".join(gSecond)+"\n"
		outfd.write(line)
		
		
def parseInparanoidTable(tableFile):
	
	geneMap=[]
	
	for line in open(tableFile):
		if line.startswith("OrtoID"): continue
		
		elems=line[:-1].split("\t")
		
		# elems[2]:source elems[3]:target
		
		#sourceOrg=elems[2][0:3]
		#targetOrg=elems[3][0:3]

		# take gene names from both lists
		# skip confidence values between genes ([::2]), drop prefix ([4:])
		#sourceGenes= [ x[4:] for x in elems[2].split()[::2] ]
		#targetGenes= [ x[4:] for x in elems[3].split()[::2] ]
		sourceGenes= [ x for x in elems[2].split()[::2] ]
		targetGenes= [ x for x in elems[3].split()[::2] ]

		geneMap.append( (sourceGenes, targetGenes) )

	return geneMap



def  parseInparanoidXML(xmlfile):
	"""parse an xml file from Inparanoid website. Returns the mapping between two organism"""
	
	ixml = parse(open(xmlfile))
	
	root = ixml.getroot()
	
	# remember all genes
	species=root.findall("species")
	
	geneLocus=dict()
	genes = defaultdict(list)
	gene2specie=dict()
	
	
	for s in species:
		specieid=s.attrib["id"]
		# specieid=s.attrib["NCBITaxId"]
		
		for gene in s.findall("database/genes/gene"):
			gid, protId = gene.attrib["id"], gene.attrib["protId"]
			genes[specieid].append(gid)
			geneLocus[gid]=protId
			gene2specie[gid]=specieid

	# remember ortholog groups
	species = genes.keys()
	clusters = root.findall("clusters/cluster")
	geneMap=[]
	
	
	for c in clusters:
		inCluster = defaultdict(list)
		for g in c.findall("geneRef"):
			gid = g.attrib["id"]
			org=gene2specie[gid]
			gName = geneLocus[gid]
			inCluster[org].append(gName)
			
		geneMap.append( [ inCluster[org] for org in species] )
		
	return species, geneMap

if __name__ == '__main__':
    main()




#!/usr/bin/env python
# encoding: utf-8
"""
untitled.py

Created by Nicolas Loira on 2010-12-18.
Copyright (c) 2010 LaBRI, Univ.Bordeaux I, France. All rights reserved.
"""

import sys
import os
# from xml.etree.ElementTree import *
import xml.etree.ElementTree as ET
from collections import defaultdict
import re
import zipfile
import logic

PTlogFD = sys.stderr

def log(m):
	PTlogFD.write(m)
	PTlogFD.write('\n')

def logRewrite(status, ga, newFormula=''):
	log("\t".join((status, ga.formula, newFormula, ga.rid, ga.getReactionName())))
# 	log( "\t".join((status, ga.formula)))


class SBMLmodel(object):
	def __init__(self, filename=None, logFD=None):

		global PTlogFD

		if logFD is not None:
			PTlogFD = logFD

		# pyversion = sys.version_info
		# if pyversion[0]>2 or (pyversion[0]==2 and pyversion[1]>=7):
		# 	# Setup ET to understand that SBML is the default namespace
		# 	ET.register_namespace('', "http://www.sbml.org/sbml/level2/version4")
		# 	ET.register_namespace('', "http://www.sbml.org/sbml/level2/version1")


		if filename:
			self.parseXML(filename)

		self.toRemove = None
		self.lockedReactions = []



	def parseXML(self, modelFile):
		assert modelFile

		if modelFile == "-":
			modelfd = sys.stdin
		else:
			modelfd = self.openModel(modelFile)

		# parse file
		result = ET.parse(modelfd)
		root = result.getroot()
		self.xmlTree = result
		self.root = root
		self.URI, self.tag = root.tag[1:].split("}", 1)
		assert self.URI, "URI was not set correctly."

		# shortcut for list of reaction nodes
		self.reacNodes = root.findall("*/{%s}listOfReactions/{%s}reaction" % (self.URI, self.URI))

		# map of all parents (useful for removing nodes)
		self.parentMap = dict((c, p) for p in root.getiterator() for c in p)

	def openModel(self, modelFile):
		assert modelFile is not None
		modelfd = None

		if zipfile.is_zipfile(modelFile):
			zp = zipfile.ZipFile(modelFile, 'r')
			firstsbml = next((name for name in zp.namelist() if name.endswith("xml") or name.endswith("sbml")), None)
			if firstsbml == None:
				log("No .xml/.sbml file found in "+modelFile)
				sys.exit(2)
			modelfd = zp.open(firstsbml, 'r')
		else:
			modelfd = open(modelFile)

		return modelfd


	def write(self, outFD):
		#self.xmlTree.write(outFD)

		# COBRA doesn't like the extra namespaces added by ElementTree,
		# so we'll parse the output
		assert self.root is not None
		assert outFD is not None

		xmlstr = ET.tostring(self.root)
		cleanxml = xmlstr

		### This would remove all ns prefixes. Handle with caution!
		# cleanxml=re.sub("<ns\d+:", "<", cleanxml)
		# cleanxml=re.sub("</ns\d+:", "</", cleanxml)
		# cleanxml=re.sub("xmlns:ns\d+=", "xmlns=", cleanxml)

		# Clean ET's ns0 default namespace, but don't touch the other ns!
		cleanxml=re.sub("<ns0:", "<", cleanxml)
		cleanxml=re.sub("</ns0:", "</", cleanxml)
		cleanxml=re.sub("xmlns:ns0=", "xmlns=", cleanxml)


		outFD.write('<?xml version="1.0" encoding="UTF-8"?>\n')
		outFD.write(cleanxml)
		outFD.write('\n')


	def deleteReactionsOutside(self, compartmentName):

		cID = None

		# find compartment id
		root = self.root
		compNodes = root.findall("*/{%s}listOfCompartments/{%s}compartment" % (self.URI, self.URI))

		for comp in compNodes:
			if comp.get('name') == compartmentName:
				cID = comp.get('id')

		# get set of metabolites in peroxisome
		speciesNodes = root.findall("*/{%s}listOfSpecies/{%s}species" % (self.URI, self.URI))

		inCompartment = set()
		for s in speciesNodes:
			if s.get('compartment') == cID:
				inCompartment.add(s.get('id'))

		# mark reactions without a foot in compartment
		if not self.toRemove:
			self.toRemove = []

		for r in self.reacNodes:
			speciesInReaction = set()
			speciesInReaction.update([sr.get('species') for sr in r.findall("{%s}listOfReactants/{%s}speciesReference" % (self.URI, self.URI))])
			speciesInReaction.update([sr.get('species') for sr in r.findall("{%s}listOfProducts/{%s}speciesReference" % (self.URI, self.URI))])
			speciesInReaction.update([sr.get('species') for sr in r.findall("{%s}listOfModifiers/{%s}modifierSpeciesReference" % (self.URI, self.URI))])

			inReactionAndCompartment = inCompartment.intersection(speciesInReaction)
			if len(inReactionAndCompartment) == 0:
				self.toRemove.append(r)

		# do the cleaning
		for reaction in self.toRemove:
			parent = self.parentMap[reaction]
			parent.remove(reaction)
			self.reacNodes.remove(reaction)

	def deleteUnusedSpecies(self):

		# remember the useful ones
		speciesInReactions = set()

		for r in self.reacNodes:
			speciesInReactions.update([sr.get('species') for sr in r.findall("{%s}listOfReactants/{%s}speciesReference" % (self.URI, self.URI))])
			speciesInReactions.update([sr.get('species') for sr in r.findall("{%s}listOfProducts/{%s}speciesReference" % (self.URI, self.URI))])
			speciesInReactions.update([sr.get('species') for sr in r.findall("{%s}listOfModifiers/{%s}modifierSpeciesReference" % (self.URI, self.URI))])

		log("species in reactions:")
		log(len(speciesInReactions))
		# remember which ones we need to delete
		toDelete = []

		speciesNodes = self.root.findall("*/{%s}listOfSpecies/{%s}species" % (self.URI, self.URI))
		log('species nodes:')
		log(len(speciesNodes))
		toDelete = [s for s in speciesNodes if s.get('id') not in speciesInReactions]

		# delete them!
		for s in toDelete:
			parent = self.parentMap[s]
			parent.remove(s)

		# done!

	def getGeneAssociations(self, reset=False):

		if not reset and hasattr(self, 'r2formula'):
			return self.r2loci, self.r2formula

		assert self.root is not None and self.URI is not None

		r2loci = dict()
		r2formula = dict()
		r2formulaNode = dict()
		rid2node = dict()

		# look for reactions

		for r in self.reacNodes:
			reacId = r.get('id', 'INVALID')
			rid2node[reacId] = r
			notes = r.find("{%s}notes" % (self.URI))
			geneFormula = None

			body = notes.find("{http://www.w3.org/1999/xhtml}body")
			if body is not None:
				notes = body


			for line in notes:
				text = line.text
				if text.startswith("GENE ASSOCIATION:") or text.startswith("GENE_ASSOCIATION:"):
					geneFormula = text[17:].strip()
					geneFormula = geneFormula.replace(" or ()","")
					geneFormula = geneFormula.replace("() or ","")
					geneFormula = geneFormula.strip()
					lineWithGA = line


			# skip reactions without gene association
			# if geneFormula is None:
			if not geneFormula:
				continue
			r2formula[reacId] = geneFormula
			r2formulaNode[reacId] = lineWithGA

			loci = frozenset(geneFormula.replace("(", "").replace(")", "").replace("and", "").replace("or", "").split())
			# $loci is now a set of locus for this reactions
			r2loci[reacId] = loci

		self.r2loci = r2loci
		self.r2formula = r2formula
		self.r2formulaNode = r2formulaNode
		self.rid2node = rid2node

		return r2loci, r2formula

	def getGAGroup(self, excludeLocked=False):
		"""generate a GAGroup object with the existing GeneAssociations"""
		r2loci, r2formula = self.getGeneAssociations()

		ridWithFormula = r2formula.keys()
		reactionsWithFormula = [r for r in self.reacNodes if r.get('id') in ridWithFormula]

		if excludeLocked:
			reactionsWithFormula = [r for r in reactionsWithFormula if r not in self.lockedReactions]

		GAs = GAGroup()
		for reac in reactionsWithFormula:
			ga = GeneAssociation(reac, self)
			GAs.add(ga)

		return GAs

	def DEPRECATED_rewriteGeneAssociations(self, rel):
		pass


	def rewriteGeneAssociations2(self, rel):
		GAs = self.getGAGroup(excludeLocked=True)
		allMapped = frozenset(rel.map1to1.keys()+rel.map1toN.keys())

		if not self.toRemove:
			self.toRemove = []

		for ga in GAs:
			newFormula = None
			if ga.loci and ga.loci <= allMapped:
				# we can translate all genes in ga.loci
				newFormula = ga.formula
				for l in ga.loci:
					lm1 = rel.map1to1.get(l, None)
					lm2 = rel.map1toN.get(l, None)
					translation = lm1 if lm1 else " or ".join(lm2)
					newFormula = newFormula.replace(l, translation)
				# log("HT Rewrite (%s) to (%s)" % (ga.formula,newFormula))
				logRewrite("HT", ga, newFormula)
			elif len(ga.loci) == 1:
				# there's no good match for this one, so take the best one we can find
				l = list(ga.loci)[0]
				bestTargets = rel.getTargetTally(frozenset(ga.loci))

				if not bestTargets:
					# try with any source group that includes l
					bestTargets = rel.getTargetTallyForLoci(l)

				if not bestTargets:
					# we really tried!
					# log("L1 no bestTargets for SG(1):(%s)" % (ga.formula))
					logRewrite("L1", ga)
					self.toRemove.append(ga)
					continue


				# choose the smallest of the bests non-empty targets
				# bestT: bestTarget

				highestTally = bestTargets[0][0]
				allCandidateTargets = [t for t in bestTargets if t[0] == highestTally and len(t[1]) > 0]
				allCandidateTargets.sort(reverse=True, key=lambda t: len(t[1]))
				bestT = allCandidateTargets[0]

				if not bestT:
					# log("F1 not even with BestQuality we could find SG(1):(%s)" % (ga.formula))
					logRewrite("F1", ga)
					continue
				else:
					# this will work fine even if there is only one elemen in bestT[1]
					translation = " or ".join(bestT[1])

					newFormula = ga.formula
					newFormula = newFormula.replace(l, translation)
					# log("B1 Rewrite (%s) to (%s)" % (ga.formula,newFormula))
					logRewrite("B1", ga, newFormula)


			else:
				# not rewritable by 1:1, or by known source group
				# try to solve easy cases first (AND/OR closure)
				nORs = ga.formula.count('or')
				nANDs = ga.formula.count('and')

				if bool(nORs) != bool(nANDs):	# (AND/OR closure)
					separator = " and " if nANDs else " or "

					if rel.inSourceGroups(ga.loci):		# easy case: sourceGroup exists
						targetGroup = rel.getBestTargetGroup(ga.loci)
						newFormula = separator.join(targetGroup)
						# log("HC Rewrite AND/OR (%s) to (%s)" % (ga.formula,newFormula))
						logRewrite("HC", ga, newFormula)
					else:
						# try to match using High and Best quality
						toTranslate = list(ga.loci)
						selectedTargets = []

						# check 1:1 HQ
						for loci in ga.loci:
							tran = rel.map1to1.get(loci, None)
							if tran:
								toTranslate.remove(loci)
								selectedTargets.append(tran)

						# get best candidates for the rest

						candidates = rel.getSourcesIncluding(toTranslate)

						# candidates is a list of tallySourceGroups: [(v, sourcegroup)]
						# sorted by votes

						# log(">> already mapped <%s>" % (str(selectedTargets)))
						# log(">> About to translate %s using table <%s>" % (str(toTranslate), str(candidates)))

						while len(toTranslate) > 0 and len(candidates) > 0:
							sourceGroup = candidates.pop(0)
							# if there is a target group, use it!
							targets = rel.getTargetTally(sourceGroup)
							if targets:
								target = targets[0]
								for s in sourceGroup[1]:
									if s in toTranslate:
										toTranslate.remove(s)

								combinedTarget = "("+ " or ".join([t for t in target[1]]) + ")"

								selectedTargets.append(combinedTarget)



						# log(">> still to translate: <%s>" % (str(toTranslate)))
						# if len(toTranslate)==0:
						if len(toTranslate) == 0 or (separator == " or " and len(selectedTargets) > 0):
							newFormula = separator.join(selectedTargets)
							# log("BC Rewrite AND/OR (%s) to (%s)" % (ga.formula,newFormula))
							logRewrite("BC", ga, newFormula)
						else:
							# log("LC No match/translation for source group: %s" % (ga.formula))
							logRewrite("LC", ga)
							self.toRemove.append(ga)

				else:
					# try to solve unsolved cases, using BestQuality map (the best we can find)


					# log("TU Unsolved case: %s" % (ga.formula))
					logRewrite("TU", ga)
					# we shouldn't, but we're going to delete them for now
					self.toRemove.append(ga)

			# rewrite the xml model
			if newFormula:

				# check for genes with pipes
				NOTPIPE = "thisisapipe"
				cleanFormula = newFormula.replace('|', NOTPIPE)

				# normalize (to CNF) and clean formula
				cleanFormula = logic.to_cnf(cleanFormula)
				cleanFormula = logic.clean_formula(cleanFormula)
				cleanFormula = logic.raise_same_op(cleanFormula)
				cleanFormula = logic.clean_formula(cleanFormula)
				cleanFormula = cleanFormula.nicerepr()

				# recover pipes
				cleanFormula = cleanFormula.replace(NOTPIPE, '|')

				log("CF\t%s\t%s" % (newFormula, cleanFormula))

				# rewrite xml
				reacId = ga.rid
				line = self.r2formulaNode[reacId]
				line.text = "GENE_ASSOCIATION: %s" % (cleanFormula)

			# next GA!

	# end of rewriting


	def forceGeneAssociations(self, forcedReactions):

		# make sure we pre-processed gene associations
		_, _ = self.getGeneAssociations()
		r2formulaNode = self.r2formulaNode

		for rid, newGA in forcedReactions.iteritems():
			log("in model.force: forcing %s to %s" % (rid, newGA))
			# TODO: this shouldn't be an assert
			assert rid in self.rid2node
			if rid in r2formulaNode:
				line = r2formulaNode[rid]
				line.text = "GENE_ASSOCIATION: %s" % (newGA)
			# mark this reaction as locked
			self.lockedReactions.append(rid)
			self.lockedReactions.append(self.rid2node[rid])


	########################################################

#			elif rel.inSourceGroups(ga.loci):
#				# let's try with ORgroups and ANDgroups
#				targetGroup=rel.getBestTargetGroup(ga.loci)
#				nORs=ga.formula.count('or')
#				nANDs=ga.formula.count('and')
#
#				if bool(nORs)!=bool(nANDs):		# logical xor
#					separator=" and " if nANDs else " or "
#					newFormula=separator.join(targetGroup)
#					log("HL Rewrite AND/OR (%s) to (%s)" % (ga.formula,newFormula))
#				else:
#					# complex case, so only log for now
#					log("TU Unsolved case: %s" % (ga.formula))
#			else:
#				# we shoulnd't reach this point
#				log("F3 No match/translation for source group: %s" % (ga.formula))


	########################################################

	def removeReactions(self, listToRemove=None):
		"""remove reactions marked as such"""

		assert not listToRemove, "Explicit list to remove: not implemented (using only reactions tagged for removal)"

		#assert self.toRemove, "Calling removeReactions before marking them"
		if not self.toRemove or len(self.toRemove) == 0:
			return

		# TODO: reference count of metabolites to remove unused entries

		for ga in self.toRemove:
			reaction = ga.reactionNode
			if reaction in self.lockedReactions:
				continue
			parent = self.parentMap[reaction]
			parent.remove(reaction)

		# reset
		self.toRemove = []


class Relations(object):

	rels = None
	name = None

	def __init__(self, filename=None):
		if filename:
			self.parseRel(filename)

	def parseRel(self, filename):
		assert filename
		self.name = os.path.basename(filename)

		rels = []
		groupMap = dict()

		for line in open(filename):
			genesOrg1, genesOrg2 = line[:-1].split('\t')

			if genesOrg1:
				genesOrg1 = frozenset(genesOrg1.split(','))
			if genesOrg2:
				genesOrg2 = frozenset(genesOrg2.split(','))

			rels.append((genesOrg1, genesOrg2))
			rels.append((genesOrg2, genesOrg1))
			groupMap[genesOrg1] = genesOrg2
			groupMap[genesOrg2] = genesOrg1

		self.rels = frozenset(rels)
		self.groupMap = groupMap

	def mappedTo(self, group):
		"""get the group this one is mapped to"""
		return self.groupMap.get(group, None)

	def __str__(self):
		elems = [str(x) for x in self.rels]
		return ";".join(elems)

class MultiRelations(object):
	"""Store and study a set of Relation mappings"""

	def __init__(self, rels=None, logFD=None):
		"""rels should be an iterable of Relations"""
		global PTlogFD
		if logFD is not None:
			PTlogFD = logFD

		self._setRels(rels)

		# init other variables
		self.votes = None
		self.map1to1 = None
		self.map1toN = None
		self.sourceToTallyGroup = None

	def debugRels(self):
		for r in self.rels:
			print r

	def _setRels(self, rels):
		self.rels = rels
		self.N = len(rels) if rels else 0

	def loadMulti(self, relFiles):
		rels = [Relations(f) for f in relFiles]
		self._setRels(rels)

	def voteForSharedGroups(self, index=0):
		assert self.rels
		relMethods = self.rels

		countGroups = defaultdict(int)
		groupToMethod = defaultdict(list)

		# to vote for methods, we are going to count the number of times each set appears
		for method in relMethods:
			for rel in method.rels:
				countGroups[rel[index]] += 1
				groupToMethod[rel[index]].append(method)

		return countGroups, groupToMethod


	def voteGroups(self):
		assert self.rels

		# already calculated
		if self.votes:
			return self.votes

		# in case of change of .rel format
		ISOURCE = 1
		ITARGET = 0

		# constructor of a defaultdict to int
		# useful to create defdict(defdict(int)), to be used as counter[A][B]+=1
		def defint():
			return defaultdict(int)

		counter = defaultdict(int)
		pairCounter = defaultdict(defint)
		mappedTo = defaultdict(set)
		votes = dict()
		sourceToTallyGroup = dict()

		# iterate over all rels, in all methods
		for method in self.rels:
			for rel in method.rels:
				sourceGroup = rel[ISOURCE]
				targetGroup = rel[ITARGET]

				# don't care about None->(targetGroup)
				if not sourceGroup:
					continue
				# neither about (sourceGroup)->None
				if not targetGroup:
					continue

				# remember mapping, count the occurrences of source and the ocurrences of source-target
				mappedTo[sourceGroup].add(targetGroup)
				counter[sourceGroup] += 1
				pairCounter[sourceGroup][targetGroup] += 1

		# vote for groups
		# (5,SG1) -> [ (3,TG2), (1,TG1), (1,TG3) ]
		for sourceGroup, sourceVotes in counter.iteritems():
			# print ">>>"+str(sourceGroup)+"<<<>>>"+str(mappedTo[sourceGroup])+"<<<"
			# print "$$$"+str(mappedTo[sourceGroup])+"$$$"
			tallySourceGroup = (sourceVotes, sourceGroup)
			#votes[tallySourceGroup] = [(pairCounter[sourceGroup][targetGroup], targetGroup) for targetGroup in mappedTo[sourceGroup]]
			votes[tallySourceGroup] = [(pairCounter[sourceGroup][tg], tg) for tg in mappedTo[sourceGroup]]

			votes[tallySourceGroup].sort(reverse=True)

			# remember this source (makes easier to look for it afterwards)
			sourceToTallyGroup[sourceGroup] = tallySourceGroup

		self.votes = votes
		self.sourceToTallyGroup = sourceToTallyGroup
		return votes

	def inSourceGroups(self, group):
		assert self.sourceToTallyGroup
		return group in self.sourceToTallyGroup


	def getBestTargetGroup(self, group):

		targets = self.getTargetTally(group)

		if targets is None:
			return None

		return targets[0][1]


	def getTargetTally(self, group):
		assert self.sourceToTallyGroup
		assert self.votes
		assert group

		if group in self.votes:
			return self.votes[group]

		st = self.sourceToTallyGroup.get(group, None)

		if st is None:
			return None

		return self.votes[st]


	def getTargetTallyForLoci(self, loci):
		assert self.sourceToTallyGroup

		allTargets = []

		for group, tallyGroup in self.sourceToTallyGroup.iteritems():
			if loci in group:
				allTargets.extend(self.votes[tallyGroup])

		# now allTargets should have a lot of target tallyPgroups!
		allTargets.sort(reverse=True)

		if len(allTargets) == 0:
			allTargets = None
		return allTargets

	def getAllSourceGroups(self):
		assert self.sourceToTallyGroup
		return self.sourceToTallyGroup.keys()

	def getSourcesIncluding(self, geneList):
		selectedSources = []
		geneSet = frozenset(geneList)

		for sourceGroup, sourceTallyGroup in self.sourceToTallyGroup.iteritems():
			commonSet = sourceGroup & geneSet
			if len(commonSet) > 0:
				selectedSources.append(sourceTallyGroup)
		selectedSources.sort(reverse=True)
		return selectedSources


	def verifyIfMappingsAgree(self, groups, groupToMethods, maxLen=99999):

		for group in groups:

			# skip groups larger than threshold
			if len(group) > maxLen:
				continue

			# get all mappings in method
			mappedTo = set()

			for method in groupToMethods[group]:
				targetGroup = method.mappedTo(group)
				if 0 < len(targetGroup) < maxLen:
					mappedTo.add(method.mappedTo(group))

			if len(mappedTo) > 1:
				log("MapMismatch: group [%s] mapped to: [%s]" % (str(group), str(mappedTo)))

	def unravel1to1(self):

		# get as many trusted 1:1 relationship between the two organisms
		assert self.N > 0
		minVotes = (self.N/2) + 1

		# get number of votes for each scaffold group (index=1), and a mapping between
		# each of those groups to the methods that produce that map
		votes = self.voteGroups()

		# store 1:1 source to target mappings (with votes!)
		map1to1 = dict()

		# get 1to1 mappings, with sourceGroups that pass the threshold, and the most voted target first!

		def goodSingle(x):
			return x[0] >= minVotes and len(x[1]) == 1

		for source, targets in votes.iteritems():
			if goodSingle(source):
				for target in targets:
					if goodSingle(target):
						# 1:1 found!
						# map1to1[source]=target
						map1to1[list(source[1])[0]] = list(target[1])[0]
						break	# go to next source
		# remember results
		self.map1to1 = map1to1

	def unravelDuplicates(self):
		"""Find 2:1, 1:2 mappings"""

		assert self.map1to1, "unravel1to1 must be called before further unraveling"

		assert self.N > 0
		minVotes = (self.N/2) + 1

		# get number of votes for each scaffold group (index=1), and a mapping between
		# each of those groups to the methods that produce that map
		votes = self.voteGroups()

		def goodGroup(x, n):
			return x[0] >= minVotes and len(x[1]) == n

		# first find 2:1 mappings
		for source, targets in votes.iteritems():
			if goodGroup(source, 2):
				for target in targets:
					if goodGroup(target, 1):
						# 2:1 found!
						self.map1to1[list(source[1])[0]] = list(target[1])[0]
						self.map1to1[list(source[1])[1]] = list(target[1])[0]
						break	# go to next source

		# now find 1:2 mappings
		if not self.map1toN:
			self.map1toN = dict()

		for source, targets in votes.iteritems():
			if goodGroup(source, 1):
				for target in targets:
					if goodGroup(target, 2):
						# 1:2 found!
						self.map1toN[list(source[1])[0]] = list(target[1])
						break	# go to next source

		# that's all
		return


class GeneAssociation(object):
	"""Store a gene association"""
	def __init__(self, reactionNode, model):
		# super(GeneAssociation, self).__init__()
		self.reactionNode = reactionNode
		self.rid = reactionNode.get('id', 'INVALID')
		self.formula = model.r2formula.get(self.rid, None)
		self.loci = model.r2loci.get(self.rid, None)

	def getReactionName(self):
		assert self.reactionNode is not None
		return self.reactionNode.get('name', 'INVALID')

class GAGroup(set):
	"""store a set of gene associations"""
	def __init__(self):
		super(GAGroup, self).__init__()


def main():
	pass

if __name__ == '__main__':
	main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""ptg_addbiggreactions: add a list bigg reactions to an existing sbml file,
making sure that the metabolites referenced by reactions are also present!"""

from __future__ import print_function

import argparse
import libsbml

from . import BiGG, SBMLutils, BiGG2SBML


__author__ = "Nicolas Loira"
__email__ = "nloira@gmail.com"
__date__ = "08/March/2016"


def drop_R(s):
	"""If exists, remove R_ prefix from reaction ids"""
	return s[2:] if s.startswith("R_") else s


def read_reaction_list(reaction_list_file):
	"""Parses a simple file with a list of reaction ids"""
	with open(reaction_list_file, 'r') as rlfd:
		reaction_list = [drop_R(l[:-1]) for l in rlfd]
	return reaction_list


def main():
	"""Add BiGG reactions to an existing SBML model"""
	# args
	parser = argparse.ArgumentParser(description="add reactions from a table, using BiGG 2.0 API")
	parser.add_argument("-s", "--sbml", required=True, help="SBML model where reactions will be added")
	parser.add_argument("-l", "--list", required=True, help="list with reactions to add")
	parser.add_argument("-o", "--output", required=True, help="output SBML file")
	args = parser.parse_args()

	sbmlfile = args.sbml
	reaction_list_file = args.list
	sbmlout = args.output

	# parse input
	BiGG.init()
	new_reactions = read_reaction_list(reaction_list_file)
	sbml = SBMLutils.read_sbml(sbmlfile)
	model = sbml.getModel()

	reactionsInModel = set([drop_R(r.getId()) for r in model.getListOfReactions()])
	reactionsAdded = set()

	# Add new reactions
	for reaction in new_reactions:
		if reaction not in reactionsAdded and reaction not in reactionsInModel:

			# try the same reactionid
			biggreaction = BiGG.get_bigg_reaction(reaction)
			if biggreaction is not None:
				BiGG2SBML.add_reaction(model, biggreaction)
				print("Adding reaction: " + biggreaction['bigg_id'])
				reactionsAdded.add(reaction)
				continue

			# else, look for old/alternative reactionid
			biggreaction = BiGG.guess_reaction(reaction)
			if biggreaction is not None:
				biggid = biggreaction['bigg_id']
				if (biggid not in reactionsAdded) and (biggid not in reactionsInModel):
					print("WARNING! replacing old reaction %s with: %s" % (reaction, biggid))
					BiGG2SBML.add_reaction(model, biggreaction)
					reactionsAdded.add(biggid)
				continue

			print("WARNING! no reaction %s in BiGG database" % (reaction))

	# close
	libsbml.writeSBMLToFile(sbml, sbmlout)
	BiGG.close()

if __name__ == '__main__':
	main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import requests
import sys
import os

from collections import defaultdict
import pprint
import time
# import cPickle as pickle
try:
	import cPickle as pickle
except:
	import pickle


"""bigglib: extract, using BiGG 2.0 API, information about reactions,
the models that contain the reactions, the associated genomes,
the associated genes, and gene association formula.
Uses an external cache file whenever possible."""

__author__ = "Nicolas Loira"
__email__ = "nloira@gmail.com"
__date__ = "14/July/2015"

BIGG_REACTION_INFO = 'http://bigg.ucsd.edu/api/v2/universal/reactions/%s'
BIGG_MODEL_INFO = 'http://bigg.ucsd.edu/api/v2/models/%s'
BIGG_REACTION_IN_MODEL = 'http://bigg.ucsd.edu/api/v2/models/%s/reactions/%s'
BIGG_METABOLITE_INFO = 'http://bigg.ucsd.edu/api/v2/universal/metabolites/%s'
BIGG_QUERY_REACTION = 'http://bigg.ucsd.edu/api/v2/search?query=%s&search_type=reactions'

BIGG_WAIT = 0.3

BIGG_COMPARTMENTS = {
	'c': 'cytosol',
	'p': 'periplasm',
	'e': 'extracellular space',
	'x': 'peroxisome',
	'm': 'mitochondria',
	'n': 'nucleus',
	'r': 'endoplasmic reticulum',
	'x': 'peroxisome/glyoxysome',
	'l': 'lysosome',
	'g': 'golgi apparatus',
	'n': 'nucleus',
	'f': 'flagellum',
	's': 'eyespot',
	'h': 'chloroplast',
	'u': 'thylakoid',
	'v': 'vacuole'}


cache = dict()
dirtycache = False
CACHE_FILE = os.path.expanduser("~/.bigg_cache.pkl")
disable_cache = False


def get_bigg_json(req):

	global cache
	global dirtycache

	if req in cache:
		return cache[req]

	results = None
	r = None

	log("Requesting to BiGG: [%s]" % req)
	try:
		r = requests.get(req)
		results = r.json()
	except Exception as e:
		if r is None:
			log("Error (%s) (%s) -- No response for request!" % (str(e), req))
		else:
			log("Error (%s) (%s) -- No json from payload:\n%s" % (str(e), req, r.text))
		results = None

	time.sleep(BIGG_WAIT)
	cache[req] = results
	dirtycache = True

	return results


def get_bigg_reaction(reactionid):

	results = get_bigg_json(BIGG_REACTION_INFO % reactionid)

	return results


def get_bigg_species(speciesid):

	results = get_bigg_json(BIGG_METABOLITE_INFO % speciesid)

	return results


def guess_reaction(reactionid):
	results = get_bigg_json(BIGG_QUERY_REACTION % reactionid)

	if results['results_count'] > 0:
		for r in results['results']:
			candidate_reaction = get_bigg_reaction(r['bigg_id'])
			if candidate_reaction is not None and 'old_identifiers' in candidate_reaction:
				if reactionid in candidate_reaction['old_identifiers']:
					return candidate_reaction

	return None


def extract_model2genomes_from_reaction(biggreaction):

	models = [m['bigg_id'] for m in biggreaction['models_containing_reaction']]

	model2genome = dict()

	for model in models:
		results2 = get_bigg_json(BIGG_MODEL_INFO % model)
		model2genome[model] = results2["genome_name"]

	return model2genome


def extract_model2ga_from_reaction(biggreaction):

	models = [m['bigg_id'] for m in biggreaction['models_containing_reaction']]

	reactionid = biggreaction['bigg_id']

	# get reaction per model
	reactionXmodel2ga = defaultdict(dict)
	reactionXmodel2genes = defaultdict(dict)

	for model in models:
		results = get_bigg_json(BIGG_REACTION_IN_MODEL % (model, reactionid))
		reactionXmodel2ga[reactionid][model] = results["results"][0]["gene_reaction_rule"]
		all_genes = results["results"][0]["genes"]
		reactionXmodel2genes[reactionid][model] = [g['bigg_id'] for g in all_genes]

	return (reactionXmodel2ga, reactionXmodel2genes)


def get_bigg_reaction_model(reactionid, modelid):
	# log("Requesting %s in model %s" % (reactionid, modelid))
	reaction_info = get_bigg_json(BIGG_REACTION_IN_MODEL % (modelid, reactionid))

	return reaction_info


def log(msg):
	"""Write to stderr"""
	sys.stderr.write(msg + '\n')


def testBIGGrm():
	r = get_bigg_reaction_model('PMI12346PH', 'iMM1415')
	pprint.pprint(r)


def testBIGG():
	x = get_bigg_reaction('PMI12346PH')
	pprint.pprint(x)

	x = get_bigg_reaction('CFAS160E')
	pprint.pprint(x)

	x = get_bigg_reaction('PMI12346PH')
	pprint.pprint(x)


def close():
	update_cache()


def update_cache():
	global cache
	global dirtycache

	if disable_cache:
		return

	if dirtycache:
		# log("/\ Updating cache...")
		with open(CACHE_FILE, 'wb') as cachefd:
			pickle.dump(cache, cachefd, -1)
		dirtycache = False
	else:
		pass
		# log("/\ Cache is clean")


def init(usecache=True):

	global disable_cache

	if usecache:
		init_cache()
	else:
		disable_cache = True


def init_cache():
	global cache

	# if there is a cache file, use it!
	if os.path.isfile(CACHE_FILE):
		with open(CACHE_FILE, 'rb') as cachefd:
			cache = pickle.load(cachefd)



#!/usr/bin/env python
# encoding: utf-8
"""
curateSBML.py

Created by Nicolas Loira on 2011.
Copyright (c) 2011 LaBRI, Univ.Bordeaux I, France. All rights reserved.
"""

import sys
import getopt
from xml.etree.ElementTree import *
import csv

help_message = '''
Patches an SBML with new reactions and species from a csv file.
Usage: curateSBML --sbml draftModel.xml -s speciesPatch.csv -r reactionPatch.csv > curatedModel.xml
'''


class Usage(Exception):
	def __init__(self, msg):
		self.msg = msg

class Species():
	def __init__(self, **kw):
		self.__dict__.update(kw)

class Reaction():
	def __init__(self, **kw):
		self.__dict__.update(kw)

	
def exiterror(errmsg=help_message):
	print >> sys.stderr, sys.argv[0].split("/")[-1] + ": " + str(errmsg)
	print >> sys.stderr, "\t for help use --help"
	sys.exit(2)

def log(msg):
	print >>sys.stderr, msg

def readcsv(filename):
	reader = csv.reader(open(filename))
	return reader

def parseSpecies(speciesFile):
	assert speciesFile
	
	allSpecies=[]
	for line in readcsv(speciesFile):
		if line[0] in ("", "ID"): continue
		allSpecies.append(Species( id=line[0], name=line[1], formula=line[2], compartment=line[3], boundary=False))

	return allSpecies


def parseReactions(reactionsFile):
	assert reactionsFile

	allReactions=[]
	for line in readcsv(reactionsFile):
		if line[0] in ("", "ID"): continue
		if line[0].startswith('#'): continue
		allReactions.append(Reaction( id=line[0], name=line[1], reversibility=line[2], reactants=line[3].split(';'), products=line[4].split(';'), geneassociation=line[5], genenameassociation=line[6], formula=line[7]))

	return allReactions

def getStoichiometry(species):
    if species.find("*")>-1:
        return species.split("*",1)
    else:
        return ("1.0", species)

class Model(object):
	def __init__(self, xmlFile):
		assert xmlFile
		model = parse(open(xmlFile))
		root=model.getroot()
		URI,tag= root.tag[1:].split("}",1)
		self.root=root
		self.URI=URI
		self.tag=tag
		self.HTMLTAG="http://www.w3.org/1999/xhtml"
		
		
		assert self.URI, "URI was not set correctly."
		
		# get a list of all known species and reactions (IDs)
		speciesNodes = root.findall("*/{%s}listOfSpecies/{%s}species" % (URI,URI))
		self.speciesIDs = [s.get('id') for s in speciesNodes]
		
		reactionNodes = root.findall("*/{%s}listOfReactions/{%s}reaction" % (URI,URI))
		self.reactionIDs = [r.get('id') for r in reactionNodes]
		
		# remember listOf... objects
		self.listOfSpecies = root.findall("*/{%s}listOfSpecies" % (URI))[0]
		self.listOfReactions = root.findall("*/{%s}listOfReactions" % (URI))[0]
		
		# map of all parents (useful for removing nodes)
		self.parentMap = dict((c, p) for p in root.getiterator() for c in p)
		
	def u(self, tag):
		return "{%s}%s" % (self.URI, tag)


	def addSpecies(self, newSpecies):
		for s in newSpecies:
			if s.id not in self.speciesIDs:
				# we'll create a new species and add it to the list of species
				newSpecies = SubElement(self.listOfSpecies, "{%s}species" % (self.URI) , id=s.id, name=s.name, compartment=s.compartment)
				if s.compartment=="boundary":
					newSpecies.attrib['boundaryCondition']="true"
			else:
				log("Warning: trying to patch existing species id: "+s.id)

	def addReactions(self, newReactions):
		for r in newReactions:
			if r.id not in self.reactionIDs:
				# we'll create a new species and add it to the list of species
				newReaction = SubElement(self.listOfReactions, "{%s}reaction" % (self.URI) , id=r.id, name=r.name, reversible=r.reversibility)
				# add notes only if needed
				if (r.geneassociation!=""):
					notes = SubElement(newReaction, "{%s}notes" % (self.URI))
					body = SubElement(notes, "{%s}body" % (self.HTMLTAG), xmlns=self.HTMLTAG)
					p1 = SubElement(body, "{%s}p" % (self.HTMLTAG) )
					p1.text="GENE_ASSOCIATION: "+r.geneassociation
				# add reactants and products
				if (len(r.reactants)>0 and r.reactants[0]!=""):
					listOfReactants = SubElement(newReaction, "{%s}listOfReactants" % (self.URI))
					for reactant in r.reactants:
						s,m=getStoichiometry(reactant)
						spref= SubElement(listOfReactants, "{%s}speciesReference" % (self.URI), species=m.strip(), stoichiometry=s )
					
				if (len(r.products)>0 and r.products[0]!=""):
					listOfProducts = SubElement(newReaction, "{%s}listOfProducts" % (self.URI))
					for product	 in r.products:
						s,m=getStoichiometry(product)
						spref= SubElement(listOfProducts, "{%s}speciesReference" % (self.URI), species=m.strip(), stoichiometry=s)
				
				# add kineticLaw
				self.addKineticLaw(newReaction, r.reversibility)
						
			# r.id already in existing reactions!
			else:
				log("Warning: trying to patch existing species id: "+s.id)

	def addKineticLaw(self, reaction, reversible="true"):
		"""
	    <kineticLaw>
          <math xmlns="http://www.w3.org/1998/Math/MathML">
            <ci> dummy_flux </ci>
          </math>
          <listOfParameters>
            <parameter id="LOWER_BOUND" value="0" units="flux_unit"/>
            <parameter id="UPPER_BOUND" value="INF" units="flux_unit"/>
            <parameter id="OBJECTIVE_COEFFICIENT" value="0" units="dimensionless"/>
          </listOfParameters>
        </kineticLaw>
		"""
		
		# add kineticLaw tags
		kineticElement=SubElement(reaction, "{%s}kineticLaw" % self.URI)
		mathElement = SubElement(kineticElement, "{%s}math" % "http://www.w3.org/1998/Math/MathML")
		ciElement = SubElement(mathElement, "{%s}ci" % "http://www.w3.org/1998/Math/MathML")
		ciElement.text= " dummy_flux "
		
		lopElement=SubElement(kineticElement, "{%s}listOfParameters" % self.URI)
		parLow=SubElement(lopElement, self.u("parameter"), id="LOWER_BOUND", units="flux_unit", value="-INF")
		parUpp=SubElement(lopElement, self.u("parameter"), id="UPPER_BOUND", units="flux_unit", value="INF")
		parObj=SubElement(lopElement, self.u("parameter"), id="OBJECTIVE_COEFFICIENT", units="dimensionless", value="0")

		if reversible in ("false", "0"):
			parLow.attrib['value']="0"
			
	def write(self):
		print '<?xml version="1.0" encoding="UTF-8"?>'
		print tostring(self.root)
	


def addSpeciesMula(xmlFile):
	assert xmlFile
	
	model = parse(open(xmlFile))
	root=model.getroot()
	URI,tag= root.tag[1:].split("}",1)
	assert URI, "URI was not set correctly."
	speciesNodes = root.findall("*/{%s}listOfSpecies/{%s}species" % (URI,URI))

	# get list of species
	speciesIds = [s.get('id') for s in speciesNodes]

	# this is the listOfSpecies object
	listOfSpecies = root.findall("*/{%s}listOfSpecies" % (URI))[0]

	toAdd=('mula1', 'mula2', 's_2973_b', 'mula4')

	for sm in toAdd:
		if sm not in speciesIds:
			# we'll create a new species and add it to the list of species
			newSpecies = SubElement(listOfSpecies, "{%s}species" % (URI) , id=sm, name="name_"+sm, compartment="c_mula")
		else:
			log("Warning: trying to patch existing species id: "+sm)
	
	# write output
	# print tostring(root)
	
	

def main(argv=None):

	sbmlfile=None
	speciesPatch=None
	reactionsPatch=None

	if argv is None:
		argv = sys.argv
	try:
		try:
			opts, args = getopt.getopt(argv[1:], "ho:vs:r:", ["help", "output=", "sbml="])
		except getopt.error, msg:
			raise Usage(msg)
	
		# option processing
		for option, value in opts:
			if option == "-v":
				verbose = True
			if option in ("-h", "--help"):
				raise Usage(help_message)
			if option in ("-o", "--output"):
				output = value
			if option=="--sbml":
				sbmlfile = value
			if option in ("-s"):
				speciesPatch = value
			if option in ("-r"):
				reactionsPatch = value

	except Usage, err:
		exiterror(err.msg)

	# validate inputs
	if not (sbmlfile and speciesPatch and reactionsPatch):
		exiterror()

	# log("sbmlfile:"+sbmlfile)
	# log("speciesfile:"+speciesPatch)
	# log("reactionsfile:"+reactionsPatch)

	# all ok, so call test method 
	model=Model(sbmlfile)

	newSpecies=parseSpecies(speciesPatch)
	newReactions=parseReactions(reactionsPatch)

	model.addSpecies(newSpecies)
	model.addReactions(newReactions)

	model.write()
	# that's all, folks!
	


if __name__ == "__main__":
	sys.exit(main())



from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='biopantograph',
      version='0.3.0',
      description='Pantograph is a toolbox for the reconstruction, curation and validation of metabolic models.',
      long_description=readme(),
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
      ],
      keywords='pantograph metabolic model reconstruction',
      url='http://pathtastic.gforge.inria.fr',
      author='Nicolas Loira',
      author_email='nloira@gmail.com',
      license='GPLv2',
      packages=['biopantograph'],
      install_requires=[ 'python-libsbml',],
      entry_points={'console_scripts': [ 
	      'ptg_addbiggreactions=biopantograph.ptg_addbiggreactions:main', 
	      'ptg_projector=biopantograph.ptg_projector:main',
	      'ptg_curateSBML=biopantograph.ptg_curateSBML:main',
	      'ptg_inparanoid2rel=biopantograph.ptg_inparanoid2rel:main',
	      'ptg_omcl2rel=biopantograph.ptg_omcl2rel:main',
	      ], },
      include_package_data=True,
      scripts=['bin/pantograph']
      )




# got it from here:
# http://aima.cs.berkeley.edu/python/logic.html
# and then I chopped it


# from __future__ import generators
# import agents
import re
import sys

sys.setrecursionlimit(3000)


#######################
# from utils import *


def find_if(predicate, seq):
    """If there is an element of seq that satisfies predicate; return it.
    >>> find_if(callable, [3, min, max])
    <built-in function min>
    >>> find_if(callable, [1, 2, 3])
    """
    for x in seq:
        if predicate(x): return x
    return None


def num_or_str(x):
    """The argument is a string; convert to a number if possible, or strip it.
    >>> num_or_str('42')
    42
    >>> num_or_str(' 42x ')
    '42x'
    """
    if isnumber(x): return x
    try:
        return int(x)
    except ValueError:
        try:
            return float(x)
        except ValueError:
                return str(x).strip()

def isnumber(x):
    "Is x a number? We say it is if it has a __int__ method."
    return hasattr(x, '__int__')


########################

class Expr:
    """A symbolic mathematical expression.  We use this class for logical
    expressions, and for terms within logical expressions. In general, an
    Expr has an op (operator) and a list of args.  The op can be:
      Null-ary (no args) op:
        A number, representing the number itself.  (e.g. Expr(42) => 42)
        A symbol, representing a variable or constant (e.g. Expr('F') => F)
      Unary (1 arg) op:
        '~', '-', representing NOT, negation (e.g. Expr('~', Expr('P')) => ~P)
      Binary (2 arg) op:
        '>>', '<<', representing forward and backward implication
        '+', '-', '*', '/', '**', representing arithmetic operators
        '<', '>', '>=', '<=', representing comparison operators
        '<=>', '^', representing logical equality and XOR
      N-ary (0 or more args) op:
        '&', '|', representing conjunction and disjunction
        A symbol, representing a function term or FOL proposition

    Exprs can be constructed with operator overloading: if x and y are Exprs,
    then so are x + y and x & y, etc.  Also, if F and x are Exprs, then so is
    F(x); it works by overloading the __call__ method of the Expr F.  Note
    that in the Expr that is created by F(x), the op is the str 'F', not the
    Expr F.   See http://www.python.org/doc/current/ref/specialnames.html
    to learn more about operator overloading in Python.

    WARNING: x == y and x != y are NOT Exprs.  The reason is that we want
    to write code that tests 'if x == y:' and if x == y were the same
    as Expr('==', x, y), then the result would always be true; not what a
    programmer would expect.  But we still need to form Exprs representing
    equalities and disequalities.  We concentrate on logical equality (or
    equivalence) and logical disequality (or XOR).  You have 3 choices:
        (1) Expr('<=>', x, y) and Expr('^', x, y)
            Note that ^ is bitwose XOR in Python (and Java and C++)
        (2) expr('x <=> y') and expr('x =/= y').
            See the doc string for the function expr.
        (3) (x % y) and (x ^ y).
            It is very ugly to have (x % y) mean (x <=> y), but we need
            SOME operator to make (2) work, and this seems the best choice.

    WARNING: if x is an Expr, then so is x + 1, because the int 1 gets
    coerced to an Expr by the constructor.  But 1 + x is an error, because
    1 doesn't know how to add an Expr.  (Adding an __radd__ method to Expr
    wouldn't help, because int.__add__ is still called first.) Therefore,
    you should use Expr(1) + x instead, or ONE + x, or expr('1 + x').
    """

    def __init__(self, op, *args):
        "Op is a string or number; args are Exprs (or are coerced to Exprs)."
        assert isinstance(op, str) or (isnumber(op) and not args)
        self.op = num_or_str(op)
        self.args = map(expr, args) ## Coerce args to Exprs

    def __call__(self, *args):
        """Self must be a symbol with no args, such as Expr('F').  Create a new
        Expr with 'F' as op and the args as arguments."""
        assert is_symbol(self.op) and not self.args
        return Expr(self.op, *args)

    def __repr__(self):
        "Show something like 'P' or 'P(x, y)', or '~P' or '(P | Q | R)'"
        if len(self.args) == 0: # Constant or proposition with arity 0
            return str(self.op)
        elif is_symbol(self.op): # Functional or Propositional operator
            return '%s(%s)' % (self.op, ', '.join(map(repr, self.args)))
        elif len(self.args) == 1: # Prefix operator
            return self.op + repr(self.args[0])
        else: # Infix operator
            return '(%s)' % (' '+self.op+' ').join(map(repr, self.args))

    def nicerepr(self):
        s=self.__repr__()
        s=s.replace('|', ' or ')
        s=s.replace('&', ' and ')
        s=s.replace('___', '-')
        s=s.replace('  ',' ')
        return s

    def __eq__(self, other):
        """x and y are equal iff their ops and args are equal."""
        return (other is self) or (isinstance(other, Expr)
            and self.op == other.op and self.args == other.args)

    def __hash__(self):
        "Need a hash method so Exprs can live in dicts."
        return hash(self.op) ^ hash(tuple(self.args))

    # See http://www.python.org/doc/current/lib/module-operator.html
    # Not implemented: not, abs, pos, concat, contains, *item, *slice
    def __lt__(self, other):     return Expr('<',  self, other)
    def __le__(self, other):     return Expr('<=', self, other)
    def __ge__(self, other):     return Expr('>=', self, other)
    def __gt__(self, other):     return Expr('>',  self, other)
    def __add__(self, other):    return Expr('+',  self, other)
    def __sub__(self, other):    return Expr('-',  self, other)
    def __and__(self, other):    return Expr('&',  self, other)
    def __div__(self, other):    return Expr('/',  self, other)
    def __truediv__(self, other):return Expr('/',  self, other)
    def __invert__(self):        return Expr('~',  self)
    def __lshift__(self, other): return Expr('<<', self, other)
    def __rshift__(self, other): return Expr('>>', self, other)
    def __mul__(self, other):    return Expr('*',  self, other)
    def __neg__(self):           return Expr('-',  self)
    def __or__(self, other):     return Expr('|',  self, other)
    def __pow__(self, other):    return Expr('**', self, other)
    def __xor__(self, other):    return Expr('^',  self, other)
    def __mod__(self, other):    return Expr('<=>',  self, other) ## (x % y)



def expr(s):
    """Create an Expr representing a logic expression by parsing the input
    string. Symbols and numbers are automatically converted to Exprs.
    In addition you can use alternative spellings of these operators:
      'x ==> y'   parses as   (x >> y)    # Implication
      'x <== y'   parses as   (x << y)    # Reverse implication
      'x <=> y'   parses as   (x % y)     # Logical equivalence
      'x =/= y'   parses as   (x ^ y)     # Logical disequality (xor)
    But BE CAREFUL; precedence of implication is wrong. expr('P & Q ==> R & S')
    is ((P & (Q >> R)) & S); so you must use expr('(P & Q) ==> (R & S)').
    >>> expr('P <=> Q(1)')
    (P <=> Q(1))
    >>> expr('P & Q | ~R(x, F(x))')
    ((P & Q) | ~R(x, F(x)))
    """
    if isinstance(s, Expr): return s
    if isnumber(s): return Expr(s)
    ## Replace the alternative spellings of operators with canonical spellings
    s = s.replace('==>', '>>').replace('<==', '<<')
    s = s.replace('<=>', '%').replace('=/=', '^')
    s = s.replace(' or ', ' | ').replace(' and ', ' & ')
    s = s.replace('-', '___')
    ## Replace a symbol or number, such as 'P' with 'Expr("P")'
    s = re.sub(r'([a-zA-Z0-9_.]+)', r'Expr("\1")', s)
    ## Now eval the string.  (A security hole; do not use with an adversary.)
    return eval(s, {'Expr':Expr})

def is_symbol(s):
    "A string s is a symbol if it starts with an alphabetic char."
    return isinstance(s, str) and s[0].isalpha()

def is_var_symbol(s):
    "A logic variable symbol is an initial-lowercase string."
    return is_symbol(s) and s[0].islower()

def is_prop_symbol(s):
    """A proposition logic symbol is an initial-uppercase string other than
    TRUE or FALSE."""
    return is_symbol(s) and s[0].isupper() and s != 'TRUE' and s != 'FALSE'


## Useful constant Exprs used in examples and code:
TRUE, FALSE, ZERO, ONE, TWO = map(Expr, ['TRUE', 'FALSE', 0, 1, 2])
A, B, C, F, G, P, Q, x, y, z  = map(Expr, 'ABCFGPQxyz')


def tt_entails(kb, alpha):
    """Use truth tables to determine if KB entails sentence alpha. [Fig. 7.10]
    >>> tt_entails(expr('P & Q'), expr('Q'))
    True
    """
    return tt_check_all(kb, alpha, prop_symbols(kb & alpha), {})

def tt_check_all(kb, alpha, symbols, model):
    "Auxiliary routine to implement tt_entails."
    if not symbols:
        if pl_true(kb, model): return pl_true(alpha, model)
        else: return True
        assert result != None
    else:
        P, rest = symbols[0], symbols[1:]
        return (tt_check_all(kb, alpha, rest, extend(model, P, True)) and
                tt_check_all(kb, alpha, rest, extend(model, P, False)))

def prop_symbols(x):
    "Return a list of all propositional symbols in x."
    if not isinstance(x, Expr):
        return []
    elif is_prop_symbol(x.op):
        return [x]
    else:
        s = set(())
        for arg in x.args:
            for symbol in prop_symbols(arg):
                s.add(symbol)
        return list(s)

def tt_true(alpha):
    """Is the sentence alpha a tautology? (alpha will be coerced to an expr.)
    >>> tt_true(expr("(P >> Q) <=> (~P | Q)"))
    True
    """
    return tt_entails(TRUE, expr(alpha))

def pl_true(exp, model={}):
    """Return True if the propositional logic expression is true in the model,
    and False if it is false. If the model does not specify the value for
    every proposition, this may return None to indicate 'not obvious';
    this may happen even when the expression is tautological."""
    op, args = exp.op, exp.args
    if exp == TRUE:
        return True
    elif exp == FALSE:
        return False
    elif is_prop_symbol(op):
        return model.get(exp)
    elif op == '~':
        p = pl_true(args[0], model)
        if p == None: return None
        else: return not p
    elif op == '|':
        result = False
        for arg in args:
            p = pl_true(arg, model)
            if p == True: return True
            if p == None: result = None
        return result
    elif op == '&':
        result = True
        for arg in args:
            p = pl_true(arg, model)
            if p == False: return False
            if p == None: result = None
        return result
    p, q = args
    if op == '>>':
        return pl_true(~p | q, model)
    elif op == '<<':
        return pl_true(p | ~q, model)
    pt = pl_true(p, model)
    if pt == None: return None
    qt = pl_true(q, model)
    if qt == None: return None
    if op == '<=>':
        return pt == qt
    elif op == '^':
        return pt != qt
    else:
        raise ValueError, "illegal operator in logic expression" + str(exp)

## NL: formula simplification and absorption law

def is_atom(s):
    return not s.args or is_symbol(s.op)

def clean_formula(s):
    """Clean redundant items (A|A), (A&A) to A"""

    if is_atom(s): return s     ## (Atoms are unchanged.)
    args=map(clean_formula, s.args)

    if s.op in ("|", "&"):
        if len(args)==2 and is_atom(args[0]) and is_atom(args[1]):
        # if both args are the same, keep only one
            if args[0]==args[1]:
                return args[0]
        else:
            # try absorption
            args=absorption_law(s, args)
            
            # remove redundancy
            present=set()
            toDelete=[]

            for a in args:
                if a in present: toDelete.append(a)
                else: present.add(a)
                
            for a in toDelete:
                args.remove(a)


    if len(args)>1:
        
        return Expr(s.op, *args)
    else:
        return args[0]



def raise_same_op(s):
    if is_atom(s): return s     ## (Atoms are unchanged.)
    
    # descend recursively
    args=map(raise_same_op, s.args)
    
    to_expand=[]
    for a in args:
        if not is_atom(a) and a.op==s.op:
            to_expand.append(a)
    
    
    for a in to_expand:
        args.extend(a.args)
        args.remove(a)
        
    return Expr(s.op, *args)
    
    


def absorption_law(s, args):
    if s.op=="|": otherop="&"
    elif s.op=="&": otherop="|"
    else: assert False, "absorption law cannot handle this op: %s" % (s.op)

    # atoms = set()
    # toDelete = []
    
    # for x in args:
    #         if is_atom(x):
    #             if x in atoms: 
    #                 toDelete.append(x) 
    #             else: 
    #                 atoms.add(x)
    # 
    #     for elem in toDelete:
    #         args.remove(elem)
    
    toDelete = set()

    atoms = [x for x in args if is_atom(x)]
    for atom in atoms:
        for elem in args:
            if not is_atom(elem) and elem.op==otherop and atom in elem.args:
                toDelete.add(elem)

    for elem in toDelete:
        while elem in args:
            args.remove(elem)
    
    return args

## Convert to Conjunctive Normal Form (CNF)

def to_cnf(s):
    """Convert a propositional logical sentence s to conjunctive normal form.
    That is, of the form ((A | ~B | ...) & (B | C | ...) & ...) [p. 215]
    >>> to_cnf("~(B|C)")
    (~B & ~C)
    >>> to_cnf("B <=> (P1|P2)")
    ((~P1 | B) & (~P2 | B) & (P1 | P2 | ~B))
    >>> to_cnf("a | (b & c) | d")
    ((b | a | d) & (c | a | d))
    >>> to_cnf("A & (B | (D & E))")
    (A & (D | B) & (E | B))
    """

    # print >>sys.stderr, s
    if isinstance(s, str): s = expr(s)
    s = eliminate_implications(s) # Steps 1, 2 from p. 215
    s = move_not_inwards(s) # Step 3
    return distribute_and_over_or(s) # Step 4

def eliminate_implications(s):
    """Change >>, <<, and <=> into &, |, and ~. That is, return an Expr
    that is equivalent to s, but has only &, |, and ~ as logical operators.
    >>> eliminate_implications(A >> (~B << C))
    ((~B | ~C) | ~A)
    """
    if not s.args or is_symbol(s.op): return s     ## (Atoms are unchanged.)
    args = map(eliminate_implications, s.args)
    a, b = args[0], args[-1]
    if s.op == '>>':
        return (b | ~a)
    elif s.op == '<<':
        return (a | ~b)
    elif s.op == '<=>':
        return (a | ~b) & (b | ~a)
    else:
        return Expr(s.op, *args)

def move_not_inwards(s):
    """Rewrite sentence s by moving negation sign inward.
    >>> move_not_inwards(~(A | B))
    (~A & ~B)
    >>> move_not_inwards(~(A & B))
    (~A | ~B)
    >>> move_not_inwards(~(~(A | ~B) | ~~C))
    ((A | ~B) & ~C)
    """
    if s.op == '~':
        NOT = lambda b: move_not_inwards(~b)
        a = s.args[0]
        if a.op == '~': return move_not_inwards(a.args[0]) # ~~A ==> A
        if a.op =='&': return NaryExpr('|', *map(NOT, a.args))
        if a.op =='|': return NaryExpr('&', *map(NOT, a.args))
        return s
    elif is_symbol(s.op) or not s.args:
        return s
    else:
        return Expr(s.op, *map(move_not_inwards, s.args))

def distribute_and_over_or(s):
    """Given a sentence s consisting of conjunctions and disjunctions
    of literals, return an equivalent sentence in CNF.
    >>> distribute_and_over_or((A & B) | C)
    ((A | C) & (B | C))
    """
    if s.op == '|':
        s = NaryExpr('|', *s.args)
        if len(s.args) == 0:
            return FALSE
        if len(s.args) == 1:
            return distribute_and_over_or(s.args[0])
        conj = find_if((lambda d: d.op == '&'), s.args)
        if not conj:
            return NaryExpr(s.op, *s.args)
        others = [a for a in s.args if a is not conj]
        if len(others) == 1:
            rest = others[0]
        else:
            rest = NaryExpr('|', *others)
        return NaryExpr('&', *map(distribute_and_over_or,
                                  [(c|rest) for c in conj.args]))
    elif s.op == '&':
        return NaryExpr('&', *map(distribute_and_over_or, s.args))
    else:
        return s

_NaryExprTable = {'&':TRUE, '|':FALSE, '+':ZERO, '*':ONE}

def NaryExpr(op, *args):
    """Create an Expr, but with an nary, associative op, so we can promote
    nested instances of the same op up to the top level.
    >>> NaryExpr('&', (A&B),(B|C),(B&C))
    (A & B & (B | C) & B & C)
    """
    arglist = []
    for arg in args:
        if arg.op == op: arglist.extend(arg.args)
        else: arglist.append(arg)
    if len(args) == 1:
        return args[0]
    elif len(args) == 0:
        return _NaryExprTable[op]
    else:
        return Expr(op, *arglist)

def conjuncts(s):
    """Return a list of the conjuncts in the sentence s.
    >>> conjuncts(A & B)
    [A, B]
    >>> conjuncts(A | B)
    [(A | B)]
    """
    if isinstance(s, Expr) and s.op == '&':
        return s.args
    else:
        return [s]

def disjuncts(s):
    """Return a list of the disjuncts in the sentence s.
    >>> disjuncts(A | B)
    [A, B]
    >>> disjuncts(A & B)
    [(A & B)]
    """
    if isinstance(s, Expr) and s.op == '|':
        return s.args
    else:
        return [s]


def pl_resolution(KB, alpha):
    "Propositional Logic Resolution: say if alpha follows from KB. [Fig. 7.12]"
    clauses = KB.clauses + conjuncts(to_cnf(~alpha))
    new = set()
    while True:
        n = len(clauses)
        pairs = [(clauses[i], clauses[j]) for i in range(n) for j in range(i+1, n)]
        for (ci, cj) in pairs:
            resolvents = pl_resolve(ci, cj)
            if FALSE in resolvents: return True
            new.union_update(set(resolvents))
        if new.issubset(set(clauses)): return False
        for c in new:
            if c not in clauses: clauses.append(c)

def pl_resolve(ci, cj):
    """Return all clauses that can be obtained by resolving clauses ci and cj.
    >>> pl_resolve(to_cnf(A|B|C), to_cnf(~B|~C|F))
    [(A | C | ~C | F), (A | B | ~B | F)]
    """
    clauses = []
    for di in disjuncts(ci):
        for dj in disjuncts(cj):
            if di == ~dj or ~di == dj:
                dnew = unique(removeall(di, disjuncts(ci)) +
                              removeall(dj, disjuncts(cj)))
                clauses.append(NaryExpr('|', *dnew))
    return clauses



def pl_fc_entails(KB, q):
    """Use forward chaining to see if a HornKB entails symbol q. [Fig. 7.14]
    >>> pl_fc_entails(Fig[7,15], expr('Q'))
    True
    """
    count = dict([(c, len(conjuncts(c.args[0]))) for c in KB.clauses
                                                 if c.op == '>>'])
    inferred = DefaultDict(False)
    agenda = [s for s in KB.clauses if is_prop_symbol(s.op)]
    if q in agenda: return True
    while agenda:
        p = agenda.pop()
        if not inferred[p]:
            inferred[p] = True
            for c in KB.clauses_with_premise(p):
                count[c] -= 1
                if count[c] == 0:
                    if c.args[1] == q: return True
                    agenda.append(c.args[1])
    return False

## Wumpus World example [Fig. 7.13]
## Fig[7,13] = expr("(B11 <=> (P12 | P21))  &  ~B11")

## Propositional Logic Forward Chanining example [Fig. 7.15]
## Fig[7,15] = PropHornKB()
##for s in "P>>Q   (L&M)>>P   (B&L)>>M   (A&P)>>L   (A&B)>>L   A   B".split():
##    Fig[7,15].tell(expr(s))


# DPLL-Satisfiable [Fig. 7.16]

def dpll_satisfiable(s):
    """Check satisfiability of a propositional sentence.
    This differs from the book code in two ways: (1) it returns a model
    rather than True when it succeeds; this is more useful. (2) The
    function find_pure_symbol is passed a list of unknown clauses, rather
    than a list of all clauses and the model; this is more efficient.
    >>> dpll_satisfiable(A&~B)
    {A: True, B: False}
    >>> dpll_satisfiable(P&~P)
    False
    """
    clauses = conjuncts(to_cnf(s))
    symbols = prop_symbols(s)
    return dpll(clauses, symbols, {})

def dpll(clauses, symbols, model):
    "See if the clauses are true in a partial model."
    unknown_clauses = [] ## clauses with an unknown truth value
    for c in clauses:
        val =  pl_true(c, model)
        if val == False:
            return False
        if val != True:
            unknown_clauses.append(c)
    if not unknown_clauses:
        return model
    P, value = find_pure_symbol(symbols, unknown_clauses)
    if P:
        return dpll(clauses, removeall(P, symbols), extend(model, P, value))
    P, value = find_unit_clause(clauses, model)
    if P:
        return dpll(clauses, removeall(P, symbols), extend(model, P, value))
    P = symbols.pop()
    return (dpll(clauses, symbols, extend(model, P, True)) or
            dpll(clauses, symbols, extend(model, P, False)))

def find_pure_symbol(symbols, unknown_clauses):
    """Find a symbol and its value if it appears only as a positive literal
    (or only as a negative) in clauses.
    >>> find_pure_symbol([A, B, C], [A|~B,~B|~C,C|A])
    (A, True)
    """
    for s in symbols:
        found_pos, found_neg = False, False
        for c in unknown_clauses:
            if not found_pos and s in disjuncts(c): found_pos = True
            if not found_neg and ~s in disjuncts(c): found_neg = True
        if found_pos != found_neg: return s, found_pos
    return None, None

def find_unit_clause(clauses, model):
    """A unit clause has only 1 variable that is not bound in the model.
    >>> find_unit_clause([A|B|C, B|~C, A|~B], {A:True})
    (B, False)
    """
    for clause in clauses:
        num_not_in_model = 0
        for literal in disjuncts(clause):
            sym = literal_symbol(literal)
            if sym not in model:
                num_not_in_model += 1
                P, value = sym, (literal.op != '~')
        if num_not_in_model == 1:
            return P, value
    return None, None


def literal_symbol(literal):
    """The symbol in this literal (without the negation).
    >>> literal_symbol(P)
    P
    >>> literal_symbol(~P)
    P
    """
    if literal.op == '~':
        return literal.args[0]
    else:
        return literal


# Walk-SAT [Fig. 7.17]

def WalkSAT(clauses, p=0.5, max_flips=10000):
    ## model is a random assignment of true/false to the symbols in clauses
    ## See ~/aima1e/print1/manual/knowledge+logic-answers.tex ???
    model = dict([(s, random.choice([True, False]))
                 for s in prop_symbols(clauses)])
    for i in range(max_flips):
        satisfied, unsatisfied = [], []
        for clause in clauses:
            if_(pl_true(clause, model), satisfied, unsatisfied).append(clause)
        if not unsatisfied: ## if model satisfies all the clauses
            return model
        clause = random.choice(unsatisfied)
        if probability(p):
            sym = random.choice(prop_symbols(clause))
        else:
            ## Flip the symbol in clause that miximizes number of sat. clauses
            raise NotImplementedError
        model[sym] = not model[sym]




def unify(x, y, s):
    """Unify expressions x,y with substitution s; return a substitution that
    would make x,y equal, or None if x,y can not unify. x and y can be
    variables (e.g. Expr('x')), constants, lists, or Exprs. [Fig. 9.1]
    >>> unify(x + y, y + C, {})
    {y: C, x: y}
    """
    if s == None:
        return None
    elif x == y:
        return s
    elif is_variable(x):
        return unify_var(x, y, s)
    elif is_variable(y):
        return unify_var(y, x, s)
    elif isinstance(x, Expr) and isinstance(y, Expr):
        return unify(x.args, y.args, unify(x.op, y.op, s))
    elif isinstance(x, str) or isinstance(y, str) or not x or not y:
        return if_(x == y, s, None)
    elif issequence(x) and issequence(y) and len(x) == len(y):
        return unify(x[1:], y[1:], unify(x[0], y[0], s))
    else:
        return None

def is_variable(x):
    "A variable is an Expr with no args and a lowercase symbol as the op."
    return isinstance(x, Expr) and not x.args and is_var_symbol(x.op)

def unify_var(var, x, s):
    if var in s:
        return unify(s[var], x, s)
    elif occur_check(var, x):
        return None
    else:
        return extend(s, var, x)

def occur_check(var, x):
    "Return true if var occurs anywhere in x."
    if var == x:
        return True
    elif isinstance(x, Expr):
        return var.op == x.op or occur_check(var, x.args)
    elif not isinstance(x, str) and issequence(x):
        for xi in x:
            if occur_check(var, xi): return True
    return False

def extend(s, var, val):
    """Copy the substitution s and extend it by setting var to val; return copy.
    >>> extend({x: 1}, y, 2)
    {y: 2, x: 1}
    """
    s2 = s.copy()
    s2[var] = val
    return s2

def subst(s, x):
    """Substitute the substitution s into the expression x.
    >>> subst({x: 42, y:0}, F(x) + y)
    (F(42) + 0)
    """
    if isinstance(x, list):
        return [subst(s, xi) for xi in x]
    elif isinstance(x, tuple):
        return tuple([subst(s, xi) for xi in x])
    elif not isinstance(x, Expr):
        return x
    elif is_var_symbol(x.op):
        return s.get(x, x)
    else:
        return Expr(x.op, *[subst(s, arg) for arg in x.args])

def fol_fc_ask(KB, alpha):
    """Inefficient forward chaining for first-order logic. [Fig. 9.3]
    KB is an FOLHornKB and alpha must be an atomic sentence."""
    while True:
        new = {}
        for r in KB.clauses:
            r1 = standardize_apart(r)
            ps, q = conjuncts(r.args[0]), r.args[1]
            raise NotImplementedError

def standardize_apart(sentence, dic):
    """Replace all the variables in sentence with new variables."""
    if not isinstance(sentence, Expr):
        return sentence
    elif is_var_symbol(sentence.op):
        if sentence in dic:
            return dic[sentence]
        else:
            standardize_apart.counter += 1
            dic[sentence] = Expr('V_%d' % standardize-apart.counter)
            return dic[sentence]
    else:
        return Expr(sentence.op, *[standardize-apart(a, dic) for a in sentence.args])

standardize_apart.counter = 0

def fol_bc_ask(KB, goals, theta):
    "A simple backward-chaining algorithm for first-order logic. [Fig. 9.6]"
    if not goals:
        yield theta
    q1 = subst(theta, goals[0])
    raise NotImplementedError


# Example application (not in the book).
# You can use the Expr class to do symbolic differentiation.  This used to be
# a part of AI; now it is considered a separate field, Symbolic Algebra.

def diff(y, x):
    """Return the symbolic derivative, dy/dx, as an Expr.
    However, you probably want to simplify the results with simp.
    >>> diff(x * x, x)
    ((x * 1) + (x * 1))
    >>> simp(diff(x * x, x))
    (2 * x)
    """
    if y == x: return ONE
    elif not y.args: return ZERO
    else:
        u, op, v = y.args[0], y.op, y.args[-1]
        if op == '+': return diff(u, x) + diff(v, x)
        elif op == '-' and len(args) == 1: return -diff(u, x)
        elif op == '-': return diff(u, x) - diff(v, x)
        elif op == '*': return u * diff(v, x) + v * diff(u, x)
        elif op == '/': return (v*diff(u, x) - u*diff(v, x)) / (v * v)
        elif op == '**' and isnumber(x.op):
            return (v * u ** (v - 1) * diff(u, x))
        elif op == '**': return (v * u ** (v - 1) * diff(u, x)
                                 + u ** v * Expr('log')(u) * diff(v, x))
        elif op == 'log': return diff(u, x) / u
        else: raise ValueError("Unknown op: %s in diff(%s, %s)" % (op, y, x))

def simp(x):
    if not x.args: return x
    args = map(simp, x.args)
    u, op, v = args[0], x.op, args[-1]
    if op == '+':
        if v == ZERO: return u
        if u == ZERO: return v
        if u == v: return TWO * u
        if u == -v or v == -u: return ZERO
    elif op == '-' and len(args) == 1:
        if u.op == '-' and len(u.args) == 1: return u.args[0] ## --y ==> y
    elif op == '-':
        if v == ZERO: return u
        if u == ZERO: return -v
        if u == v: return ZERO
        if u == -v or v == -u: return ZERO
    elif op == '*':
        if u == ZERO or v == ZERO: return ZERO
        if u == ONE: return v
        if v == ONE: return u
        if u == v: return u ** 2
    elif op == '/':
        if u == ZERO: return ZERO
        if v == ZERO: return Expr('Undefined')
        if u == v: return ONE
        if u == -v or v == -u: return ZERO
    elif op == '**':
        if u == ZERO: return ZERO
        if v == ZERO: return ONE
        if u == ONE: return ONE
        if v == ONE: return u
    elif op == 'log':
        if u == ONE: return ZERO
    else: raise ValueError("Unknown op: " + op)
    ## If we fall through to here, we can not simplify further
    return Expr(op, *args)

def d(y, x):
    "Differentiate and then simplify."
    return simp(diff(y, x))
